-- (c) M.P.I 1998/99-2001/02 -- Last Update: 02.02.03

module RList where

import Mpi

-- (1) Datatype definition -----------------------------------------------------

data RList a = Nil | Cons (a , RList a) deriving Show

inRList = either (const Nil) Cons

outRList Nil         = Left ()
outRList (Cons(a,x)) = Right(a,x)

-- (2) Ana + cata + hylo -------------------------------------------------------

cataRList g   = g . recRList (cataRList g) . outRList   

anaRList g    = inRList . (recRList (anaRList g) ) . g

hyloRList h g = cataRList h . anaRList g

baseRList g f = id -|- (g >< f)

recRList f    = id -|- (id >< f)

-- (3) Map ---------------------------------------------------------------------

instance Functor RList
         where fmap f = cataRList ( inRList . (baseRList f id ))

-- (4) Examples ----------------------------------------------------------------

-- (4.1) Square function -------------------------------------------------------

sq = hyloRList summing odds

summing = either (const 0) (uncurry (+))

odds 0 = i1 ()
odds (n+1) = i2 (2*n+1,n)

{-- pointwise:
sq 0 = 0
sq (n+1) = 2*n+1 + sq n
--}

-- (4.2) Factorial--------------------------------------------------------------

fac = hyloRList (either (const 1) mul) nats

-- where

mul(x,y) = x*y

nats = (id -|- (split succ id)) . outNat

outNat 0 = Left ()
outNat (n+1) = Right n

-- (4.2.1) Factorial (alternative) ---------------------------------------------

fac' = hyloRList (either (const 1) (mul . (succ >< id)))
                 ((id -|- (split id id)) . outNat)

{-- cf:

fac' = hyloRList (either (const 1) g) nats'
       where g(n,m) = (n+1) * m
             nats' 0 = Left ()
             nats' (n+1) = Right (n,n)
--}

-- (4.3) Insertion sort --------------------------------------------------------

iSort :: Ord a => [a] -> [a]
iSort = hyloRList (either (const []) insert) conv

-- where

conv []    = Left  ()
conv (h:t) = Right (h,t)

insert(x,[])              = [x]
insert(x,a:l) | x < a     = [x,a]++l
              | otherwise = a:(insert(x,l))

-- (4.4) converting RList to haskell lists -------------------------------------

cRList2h = cataRList (either (const []) (uncurry (:)))

ch2RList = foldr (curry Cons) Nil

-- (4.5) RList inversion -------------------------------------------------------

invl = cataRList (either (const Nil) (conc . swap . (singl >< id)))
       where singl a = Cons(a,Nil)
	     conc(l,r) = ch2RList ((cRList2h l) ++ (cRList2h r))

-- (4.6) Look-up function ------------------------------------------------------

look :: Eq a => a -> RList (a,b) -> Maybe b
look k = cataRList (either (const Nothing) aux)
         where aux((a,b),r)
                      | a == k    = Just b
                      | otherwise = r

--------------------------------------------------------------------------------

