
import Mpi

data StTra s a = T { the :: (s -> (a, s)) }

instance Functor (StTra s) where
         fmap f t = do { a <- t ; return (f a) }
-- ie:   fmap f (T g) = T(\s -> let (a,s') = g s in (f a,s'))

instance Monad (StTra s) where
         return a    = T (\x -> (a,x))
         (T x) >>= f = T (ap . ( (the . f) >< id) . x)
-- ie:   (T f) >>= g = T (\s -> let (a,s') = f s
--                                  T k    = g a
--                              in k s')

-- program execution

exec (T prog) s = prog s

-- generic state transformers 

update :: (a -> a) -> StTra a a
update f = T(split id f)

set    :: a  -> StTra a a
set a = update (const a)

-- mobile phone directory

store n = T((split length id) . npush10 n)

del n = T((split length id) . tail . npush10 n)

call n = T((split head id) . npush10 n)

npush10 :: Eq a => a -> [a] -> [a]
npush10 n = (take 10) . (npush n)
            where npush n s = [n] ++ (filter (/=n) s)

storebatch l = sequence_ (do i <- l ; return (store i))
                          -- cf [ store i | i <- l ] 

clear = update (const [])

{-- test suite

exec (do { clear ; store 253000111 ; store 253000222 ; call 253000333 ; del 253000111 } ) [2]

exec (sequence_ [ store i | i <- [1,2,3]] ) [4]

--}

