
/* implementacao em C de 'undistr' -- pointwise e quasi-pointfree     */
/* este exemplo ilustra a utilizacao (limitada) em C de conceitos     */
/* bem conhecidos no mundo funcional - programacao de ordem superior  */
/* e polimorfismo                                                      */

#include <stdlib.h>

/* declaracao de um tipo produto (nao-polimorfico) A X B */

typedef struct {
  int first;
  char second[10];
} int_X_string;

/* e suas projeccoes */

int* p1_int_X_string (int_X_string *arg)
{
  int *new = malloc(sizeof(int));
  *new = arg->first;
  return new;
}

char* p2_int_X_string (int_X_string *arg)
{
  char *new = malloc(sizeof(char) * strlen(arg->second) + 1);
  strcpy(new, arg->second);
  return new;
}

/* tipo produto A X C */

typedef struct {
  int first;
  float second;
} int_X_float;

int* p1_int_X_float (int_X_float *arg)
{
  int *new = malloc(sizeof(int));
  *new = arg->first;
  return new;
}

float* p2_int_X_float (int_X_float *arg)
{
  float *new = malloc(sizeof(float));
  *new = arg->second;
  return new;
}

/* declaracao de um tipo soma - (A X B) + (A X C) */

typedef struct {
  int i;
  union {
    int_X_string ifi1;
    int_X_float ifi2;
  } data;
} iXs_plus_iXf;

/* suas injeccoes */

iXs_plus_iXf* i1_iXs_plus_iXf (int_X_string *arg)
{
  iXs_plus_iXf *new = malloc(sizeof(iXs_plus_iXf));
  new->i = 1;
  new->data.ifi1 = *arg;
  return new;
}

iXs_plus_iXf* i2_iXs_plus_iXf (int_X_float *arg)
{
  iXs_plus_iXf *new = malloc(sizeof(iXs_plus_iXf));
  new->i = 2;
  new->data.ifi2 = *arg;
  return new;
}

/*  tipo soma (B + C) */

typedef struct {
  int i;
  union {
    char ifi1[10];
    float ifi2;
  } data;
} string_plus_float;

string_plus_float* i1_string_plus_float (char* arg)
{
  string_plus_float *new = malloc(sizeof(string_plus_float));
  new->i = 1;
  strcpy(arg, new->data.ifi1);
  return new;
}

string_plus_float* i2_string_plus_float (float* arg)
{
  string_plus_float *new = malloc(sizeof(string_plus_float));
  new->i = 2;
  new->data.ifi2 = *arg;
  return new;
}

/* tipo produto A X (B + C) */

typedef struct {
  int first;
  string_plus_float second;
} int_X_spf;

int* p1_int_X_spf (int_X_spf *arg)
{
  int *new = malloc(sizeof(int));
  *new = arg->first;
  return new;
}

string_plus_float* p2_int_X_spf (int_X_spf *arg)
{
  string_plus_float *new = malloc(sizeof(string_plus_float));
  *new = arg->second;
  return new;
}


/* implementacao de uma funcao split -- calcula o par (f arg, g arg)  */
/* nao devolve funcao, mas sim o par calculado!                       */
/* E' especifica do produto A X (B + C) mas polimorfica no tipo de arg  */

int_X_spf* split (int* (*f) (void *), string_plus_float* (*g) (void *), 
		  void *arg)
{
  int_X_spf* new = malloc(sizeof(int_X_spf));
  new -> first = *(*f)(arg);
  new -> second = *(*g)(arg);
  return new;
}


/* funcao either especifica do tipo soma (A X B) + (A X C) */
/* polimorfica no tipoi de arg                             */

void* either (void*(*f)(int_X_string*), void*(*g)(int_X_float*), 
	      iXs_plus_iXf *arg)
{
  if (arg->i == 1) return (*f)(&(arg->data.ifi1));
  else return (*g)(&(arg->data.ifi2));
}


/* sege-se a construcao, passo a passo, de undistr. As funcoes     */
/* argumento do plit final sao forcosamente construidas point-wise! */
  
int* aa (iXs_plus_iXf* arg)
{
  return either ((void*(*)(int_X_string*))p1_int_X_string, 
		 (void*(*)(int_X_float*))p1_int_X_float, 
		 arg);
}

string_plus_float* b1 (int_X_string  *arg)
{
  return i1_string_plus_float(p2_int_X_string(arg));
}

string_plus_float* b2 (int_X_float *arg)
{
  return i2_string_plus_float(p2_int_X_float(arg));
}

string_plus_float* bb (iXs_plus_iXf *arg)
{
  return either ((void*(*)(int_X_string*))b1, 
		 (void*(*)(int_X_float*))b2, 
		 arg);
}

/* construcao final -- sao omitidos os casts dos tipos */

int_X_spf* undistr(iXs_plus_iXf *arg)
{
  return split (aa, bb, arg);
}


/* versao point-wise da mesma funcao */

int_X_spf* undistr2(iXs_plus_iXf *arg)
{
  int_X_spf* new = malloc(sizeof(int_X_spf));
  switch (arg->i) {
  case 1: 
    new->first = arg->data.ifi1.first;
    strcpy(new->second.data.ifi1,arg->data.ifi1.second);
    new->second.i = 1;
    break;
  case 2:
    new->first = arg->data.ifi2.first;
    new->second.data.ifi2 = arg->data.ifi2.second;
    new->second.i = 2;
    break;
  }
  return new;
}
